function [xyz,f,freq,flag] = BFGS_trust_region(xyz,tol,iter_max)
global rstar feval
% returns flag = 1 only if converged to a minimum
%    Rmax = 0.1; % the maximal trust region radius

rstar = 2^(1/6); % argument of the minimum of the Lennard-Jones pair potential V(r) = 6^(-12) - r^(-6)
% column i corresponds to atom i, row 1 -> x, row 2 -> y, row 3 -> z

% fprintf('In BFGS\n');
Rmax = 0.1;
[m Na] = size(xyz);
xyz1 = xyz;
    %% Set the initial vector 
x = remove_rotations_translations(xyz);
N = length(x);

    
    [f, g] = LJ(x);
    nor = norm(g);
%    Rmax = 0.1; % the maximal trust region radius
    R = Rmax/5;;
    Rmin = 1e-14;
    rho_good = 0.75;
    rho_bad = 0.25;
    eta = 0.01;
    
    n_reset = 20; % reset hessian to I every n_reset iterations

%     fprintf('Initially, V = %d\t, |grad V| = %d\n',f,nor); 
    %% The trust region BFGS method
    tic
    
    iter = 0;
    feval = 0;
    flag = 0;
    I = eye(length(x));
    B = I;
    while nor > tol & iter < iter_max
        % solve the constrained minimization problem using dogleg strategy
        if cond(B) > 1e12
            B = I;
        end
        pu = -g*(g'*g)/(g'*B*g);
        pb = -B\g;
        normpu = norm(pu);
        normpb = norm(pb);
        if normpu > R 
            p = R*pu/normpu;
        else
            if normpb < R
                p = pb;
            else
                d = pb - pu;
                root = roots([d'*d, 2*pu'*d, pu'*pu - R^2]);
                if sum(isfinite(root)) < 2
                    flag = -1;
                else
                    alp = max(roots([d'*d, 2*pu'*d, pu'*pu - R^2]));                
                    p = pu + alp*d;
                end
            end
        end
        iter = iter + 1;  
        if flag == -1
            break;
        end
        
        % assess the progress
        xnew = x + p;
        [fnew, gnew] = LJ(xnew);
        mnew = f + g'*p + 0.5*p'*B*p;
%        fprintf('f = %d\tfnew = %d\t mnew = %d\n',f,fnew,mnew);
        rho = (f - fnew+1e-14)/(f - mnew+1e-14);
        % adjust the trust region
        if rho < rho_bad
            R = max([0.25*R,Rmin]);
        else
            if rho > rho_good & abs(norm(p) - R) < tol
                R = min([Rmax,2*R]);
            end
        end
        % accept or reject step
        if rho > eta            
            s = xnew - x;
            y = gnew - g;
            aux = B*s;
            B = B - (aux*aux')/(s'*aux) + y*y'/(y'*s);
            x = xnew;
            f = fnew;
            g = gnew;
            nor = norm(g);
%            fprintf('Accept: iter = %d, rho = %d\t f = %d, |grad f| = %d, R = %d\n',iter,rho,f,nor,R);
        else
%            fprintf('Reject: iter = %d,rho = %d\n',iter,rho);
        end
%         if mod(iter,n_reset) == 0
%             B = I;
%         end
    end
    if iter == iter_max
        flag = -1;
    end
    cputime = toc;
    %% check what is found
    if flag == 0
%         fprintf('Result: f = %.8f\t |df| = %d\t niter = %d\t feval = %d\n',f,nor,iter,feval);
        H = LJhess(x);
        ihess = sort(eig(H),'ascend');
        freq = sum(log(abs(ihess)));
        if ihess(1) > 0 % if a nondegenerate minimum is found
            flag = 1;
            MorseIndex = 0;
        else
            flag = -1;
        end
        x_aux = [0;0;0;x(1);0;0;x(2:3);0;x(4:N)];
        xyz = reshape(x_aux,3,Na);
    else
        fprintf('BFGS trust region method failed to find a minimum\n');
        xyz = xyz1;
        freq = 0;
        flag = 0;
    end
        

end
  %% Lennard-Jones potential        
function [v dv] = LJ(xyz) % 
global feval
feval = feval + 1;
[m n] = size(xyz);
Na = (m + 6)/3 ;
x_aux = [0;0;0;xyz(1);0;0;xyz(2:3);0;xyz(4:m)];
m = length(x_aux);
% restore atomic coordinates 
x = reshape(x_aux,3,Na);
r = zeros(Na);
for k = 1 : Na
    r(k,:) = sqrt(sum((x - x(:,k)*ones(1,Na)).^2,1));
end
r = r + diag(ones(Na,1));
aux = 1./r.^6;
L = (aux - 1).*aux;
L = L - diag(diag(L));
v = 2*sum(sum(L));

g = zeros(size(x_aux));
L = -6*(2./r.^6 - 1)./r.^8;
for k = 1 : Na
    Lk = L(:,k);
    g(1 + (k-1)*3 : k*3) = (x(:,k)*ones(1,Na) - x)*Lk;
end

g = 4*g;
dv = g([4,7,8,10 : Na*3]);

end

%%
function v = LJpot(xyz) % 
[m n] = size(xyz);
Na = (m + 6)/3 ;
x_aux = [0;0;0;xyz(1);0;0;xyz(2:3);0;xyz(4:m)];
m = length(x_aux);
% restore atomic coordinates 
x = reshape(x_aux,3,Na);
r = zeros(Na);
for k = 1 : Na
    r(k,:) = sqrt(sum((x - x(:,k)*ones(1,Na)).^2,1));
end
r = r + diag(ones(Na,1));
aux = 1./r.^6;
L = (aux - 1).*aux;
L = L - diag(diag(L));
v = 2*sum(sum(L));
end
 
 %%
function dv = LJgrad(xyz) % 
[m n] = size(xyz);
Na = (m + 6)/3 ;
x_aux = [0;0;0;xyz(1);0;0;xyz(2:3);0;xyz(4:m)];
m = length(x_aux);
% restore atomic coordinates
g = zeros(size(x_aux));
x = reshape(x_aux,3,Na);
r = zeros(Na);
for k = 1 : Na
    r(k,:) = sqrt(sum((x - x(:,k)*ones(1,Na)).^2,1));
end
r = r + diag(ones(Na,1));
L = -6*(2./r.^6 - 1)./r.^8;
for k = 1 : Na
    Lk = L(:,k);
    g(1 + (k-1)*3 : k*3) = (x(:,k)*ones(1,Na) - x)*Lk;
end

g = 4*g;
dv = g([4,7,8,10 : Na*3]);
end

%%
 function dv = LJder(r)
 dv = -12*r.^(-13) + 6*r.^(-7);
 end
%%
%%
function H = LJhess(x) % find the Hessian using finite differences
h = 1e-4;
n = length(x);
H = zeros(n);
e = eye(n);
for i = 1 : n
    di = e(:,i)*h;
    Hei = 0.5*(LJgrad(x + di) - LJgrad(x - di))/h;
    for j = 1 : i
        H(j,i) = e(j,:)*Hei;
        H(i,j) = H(j,i);
    end
end
end

